/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/logging.h"
#include "boost/core/swap.hpp" // for swap
#include "boost/iterator/iterator_facade.hpp"
#include "boost/log/attributes/clock.hpp"
#include "boost/log/core/core.hpp" // for core
#include "boost/log/core/record.hpp"
#include "boost/log/sources/record_ostream.hpp"
#include "boost/log/utility/setup/from_settings.hpp"
#include "boost/log/utility/setup/settings.hpp"
#include "boost/move/utility_core.hpp" // for move
#include "boost/multi_index/detail/bidir_node_iterator.hpp"
#include "boost/operators.hpp"
#include "boost/preprocessor/seq/limits/enum_256.hpp"
#include "boost/preprocessor/seq/limits/size_256.hpp"
#include "boost/property_tree/detail/exception_implementation.hpp"
#include "boost/smart_ptr/shared_ptr.hpp"
#include "boost/tuple/detail/tuple_basic.hpp" // for get
#include "caosdb/log_level.h"
#include <memory>
#include <sstream>
#include <string>
#include <utility> // for move
#include <vector>

namespace caosdb::logging {

BOOST_LOG_GLOBAL_LOGGER_INIT(logger, boost_logger_class) {
  boost_logger_class lg;
  return lg;
}
LoggingConfiguration::LoggingConfiguration(int level) : LevelConfiguration(level) {}

auto LoggingConfiguration::AddSink(const std::shared_ptr<SinkConfiguration> &sink) -> void {
  this->sinks.push_back(sink);
}

auto LoggingConfiguration::GetSinks() const
  -> const std::vector<std::shared_ptr<SinkConfiguration>> & {
  return this->sinks;
}

SinkConfiguration::SinkConfiguration(std::string name, int level)
  : LevelConfiguration(level), name(std::move(name)) {}

[[nodiscard]] auto SinkConfiguration::GetName() const -> const std::string & { return this->name; }

auto SinkConfiguration::Configure(boost::log::settings &settings) const -> void {
  CAOSDB_LOG_TRACE(logger_name) << "Enter SinkConfiguration::Configure(&settings)";
  auto sink = "Sinks." + GetName();
  settings[sink]["Destination"] = GetDestination();
  settings[sink]["Filter"] = "%Severity% >= " + std::to_string(GetLevel());
  settings[sink]["AutoFlush"] = true;
  settings[sink]["Format"] = "[%TimeStamp%][%Severity%] %Channel% - %Message%";
}

ConsoleSinkConfiguration::ConsoleSinkConfiguration(const std::string &name, int level)
  : SinkConfiguration(name, level) {}

[[nodiscard]] auto ConsoleSinkConfiguration::GetDestination() const -> const std::string & {
  CAOSDB_LOG_TRACE(logger_name) << "Enter ConsoleSinkConfiguration::GetDestination()";
  return this->destination;
}

auto ConsoleSinkConfiguration::Configure(boost::log::settings &settings) const -> void {
  CAOSDB_LOG_TRACE(logger_name) << "Enter ConsoleSinkConfiguration::Configure(&settings)";
  sink_configuration::Configure(settings);
}

FileSinkConfiguration::FileSinkConfiguration(const std::string &name, int level)
  : SinkConfiguration(name, level) {}

[[nodiscard]] auto FileSinkConfiguration::GetDestination() const -> const std::string & {
  CAOSDB_LOG_TRACE(logger_name) << "Enter FileSinkConfiguration::GetDestination()";
  return this->destination;
}

auto FileSinkConfiguration::SetDirectory(const std::string &directory) -> void {
  this->directory = std::string(directory);
}

auto FileSinkConfiguration::Configure(boost::log::settings &settings) const -> void {
  CAOSDB_LOG_TRACE(logger_name) << "Enter FileSinkConfiguration::Configure(&settings)";
  sink_configuration::Configure(settings);
  settings["Sink." + GetName() + ".Target"] = this->directory;
}

SyslogSinkConfiguration::SyslogSinkConfiguration(const std::string &name, int level)
  : SinkConfiguration(name, level) {}

[[nodiscard]] auto SyslogSinkConfiguration::GetDestination() const -> const std::string & {
  return this->destination;
}

// Called if no custom logging settings are specified.
auto initialize_logging_defaults() -> int {
  // first: turn everything off
  boost::log::settings off_settings;
  off_settings["Core.DisableLogging"] = true;
  boost::log::init_from_settings(off_settings);

  // now set everything up
  const static std::vector<std::shared_ptr<SinkConfiguration>> default_sinks = {
    std::make_shared<ConsoleSinkConfiguration>("DEFAULT_SINK_1", CAOSDB_DEFAULT_LOG_LEVEL)};

  boost::log::settings default_settings;

  default_settings["Core.DisableLogging"] = false;

  for (const auto &sink : default_sinks) {
    sink->Configure(default_settings);
  }

  boost::log::init_from_settings(default_settings);
  auto core = boost::log::core::get();
  core->add_global_attribute("TimeStamp", boost::log::attributes::local_clock());

  CAOSDB_LOG_DEBUG(logger_name) << "Initialized default settings.";

  return 0;
}

// Called if custom logging settings are specified.
auto initialize_logging(const LoggingConfiguration &configuration) -> void {
  // first: turn everything off
  boost::log::settings off_settings;
  off_settings["Core.DisableLogging"] = true;
  boost::log::init_from_settings(off_settings);

  // now set everything up
  boost::log::settings new_settings;

  if (configuration.GetLevel() == CAOSDB_LOG_LEVEL_OFF) {
    new_settings["Core.DisableLogging"] = true;
    return;
  } else {
    new_settings["Core.DisableLogging"] = false;
  }

  for (const auto &sink : configuration.GetSinks()) {
    sink->Configure(new_settings);
  }

  boost::log::init_from_settings(new_settings);

  CAOSDB_LOG_DEBUG(logger_name) << "Initialized logging with custom settings.";
}

void caosdb_log_fatal(const char *channel, const char *msg) {
  BOOST_LOG_CHANNEL_SEV(caosdb::logging::logger::get(), channel, CAOSDB_LOG_LEVEL_FATAL) << msg;
}

void caosdb_log_error(const char *channel, const char *msg) {
  BOOST_LOG_CHANNEL_SEV(caosdb::logging::logger::get(), channel, CAOSDB_LOG_LEVEL_ERROR) << msg;
}

void caosdb_log_warn(const char *channel, const char *msg) {
  BOOST_LOG_CHANNEL_SEV(caosdb::logging::logger::get(), channel, CAOSDB_LOG_LEVEL_WARN) << msg;
}

void caosdb_log_info(const char *channel, const char *msg) {
  BOOST_LOG_CHANNEL_SEV(caosdb::logging::logger::get(), channel, CAOSDB_LOG_LEVEL_INFO) << msg;
}

void caosdb_log_debug(const char *channel, const char *msg) {
  BOOST_LOG_CHANNEL_SEV(caosdb::logging::logger::get(), channel, CAOSDB_LOG_LEVEL_DEBUG) << msg;
}

void caosdb_log_trace(const char *channel, const char *msg) {
  BOOST_LOG_CHANNEL_SEV(caosdb::logging::logger::get(), channel, CAOSDB_LOG_LEVEL_TRACE) << msg;
}

} // namespace caosdb::logging
