/*
 * This file is a part of the CaosDB Project.
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 Florian Spreckelsen <f.spreckelsen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program. If not, see
 * <https://www.gnu.org/licenses/>.
 *
 */
#include "ccaosdb.h"
#include "caosdb/connection.h"
#include "caosdb/constants.h"
#include "caosdb/data_type.h" // for DataType, AtomicDat...
#include "caosdb/utility.h"
#include "caosdb/status_code.h"
#include "caosdb/logging.h"
#include <cassert>
#include <cstring>
#include <exception>
#include <iostream>
#include <stdio.h>
#include <string.h>
#include <vector>

extern "C" {

#define CCAOSDB_LOGGER_NAME "ccaosdb"

#define WRAPPED_ENTITY_CAST(name) static_cast<caosdb::entity::Entity *>(name->wrapped_entity)

#define WRAPPED_PROPERTY_CAST(name) static_cast<caosdb::entity::Property *>(name->wrapped_property)

#define WRAPPED_PARENT_CAST(name) static_cast<caosdb::entity::Parent *>(name->wrapped_parent)

#define WRAPPED_MESSAGE_CAST(name) static_cast<caosdb::entity::Message *>(name->wrapped_message)

#define ENUM_NAME_FROM_VALUE(arg, etype)                                                           \
  caosdb::utility::getEnumNameFromValue<caosdb::entity::etype>(arg)

#define ENUM_VALUE_FROM_NAME(arg, etype)                                                           \
  caosdb::utility::getEnumValueFromName<caosdb::entity::etype>(arg)

/*
 * Macro for wrapping every function into a try-catch clause. If an exception
 * occurs, the given StatusCode is being returned.
 */
#define ERROR_RETURN_CODE(code, fun, body)                                                         \
  fun {                                                                                            \
    CAOSDB_LOG_TRACE(CCAOSDB_LOGGER_NAME) << "Enter " << #fun;                                     \
    try {                                                                                          \
      body                                                                                         \
    } catch (const std::exception &exc) {                                                          \
      caosdb::logging::caosdb_log_fatal(CCAOSDB_LOGGER_NAME, exc.what());                          \
      return caosdb::StatusCode::code;                                                             \
    }                                                                                              \
  }

/**
 * Macro for entity getters
 */
#define CAOSDB_ENTITY_GET(element, GetFunction)                                                    \
  ERROR_RETURN_CODE(                                                                               \
    GENERIC_ERROR,                                                                                 \
    int caosdb_entity_entity_get_##element(caosdb_entity_entity *entity, char **out), {            \
      auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);                                          \
      auto *tmp = (char *)malloc(sizeof(char) * wrapped_entity->GetFunction.length() + 1);         \
      strcpy(tmp, wrapped_entity->GetFunction.c_str());                                            \
      delete[] * out;                                                                              \
      *out = tmp;                                                                                  \
      return 0;                                                                                    \
    })

/**
 * Macro for entity setters
 */
#define CAOSDB_ENTITY_SET(element, value, body_part)                                               \
  ERROR_RETURN_CODE(                                                                               \
    GENERIC_ERROR,                                                                                 \
    int caosdb_entity_entity_set_##element(caosdb_entity_entity *entity, const char *value), {     \
      auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);                                          \
      body_part return 0;                                                                          \
    })

/**
 * Macro for property getters
 */
#define CAOSDB_PROPERTY_GET(element, GetFunction)                                                  \
  ERROR_RETURN_CODE(                                                                               \
    GENERIC_ERROR,                                                                                 \
    int caosdb_entity_property_get_##element(caosdb_entity_property *property, char **out), {      \
      auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);                                    \
      auto *tmp = (char *)malloc(sizeof(char) * wrapped_property->GetFunction.length() + 1);       \
      strcpy(tmp, wrapped_property->GetFunction.c_str());                                          \
      delete[] * out;                                                                              \
      *out = tmp;                                                                                  \
      return 0;                                                                                    \
    })

/**
 * Macro for property setters
 */
#define CAOSDB_PROPERTY_SET(element, value, body_part)                                             \
  ERROR_RETURN_CODE(                                                                               \
    GENERIC_ERROR,                                                                                 \
    int caosdb_entity_property_set_##element(caosdb_entity_property *property, const char *value), \
    {                                                                                              \
      auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);                                    \
      body_part return 0;                                                                          \
    })

/**
 * Macro for parent getters
 */
#define CAOSDB_PARENT_GET(element, GetFunction)                                                    \
  ERROR_RETURN_CODE(                                                                               \
    GENERIC_ERROR,                                                                                 \
    int caosdb_entity_parent_get_##element(caosdb_entity_parent *parent, char **out), {            \
      auto *wrapped_parent = WRAPPED_PARENT_CAST(parent);                                          \
      auto *tmp = (char *)malloc(sizeof(char) * wrapped_parent->GetFunction.length() + 1);         \
      strcpy(tmp, wrapped_parent->GetFunction.c_str());                                            \
      delete[] * out;                                                                              \
      *out = tmp;                                                                                  \
      return 0;                                                                                    \
    })

/**
 * Macro for parent setters
 */
#define CAOSDB_PARENT_SET(element, value, body_part)                                               \
  ERROR_RETURN_CODE(                                                                               \
    GENERIC_ERROR,                                                                                 \
    int caosdb_entity_parent_set_##element(caosdb_entity_parent *parent, const char *value), {     \
      auto *wrapped_parent = WRAPPED_PARENT_CAST(parent);                                          \
      body_part return 0;                                                                          \
    })

int caosdb_constants_LIBCAOSDB_VERSION_MAJOR() { return caosdb::LIBCAOSDB_VERSION_MAJOR; }

int caosdb_constants_LIBCAOSDB_VERSION_MINOR() { return caosdb::LIBCAOSDB_VERSION_MINOR; }

int caosdb_constants_LIBCAOSDB_VERSION_PATCH() { return caosdb::LIBCAOSDB_VERSION_PATCH; }

int caosdb_constants_COMPATIBLE_SERVER_VERSION_MAJOR() {
  return caosdb::COMPATIBLE_SERVER_VERSION_MAJOR;
}

int caosdb_constants_COMPATIBLE_SERVER_VERSION_MINOR() {
  return caosdb::COMPATIBLE_SERVER_VERSION_MINOR;
}

int caosdb_constants_COMPATIBLE_SERVER_VERSION_PATCH() {
  return caosdb::COMPATIBLE_SERVER_VERSION_PATCH;
}

const char *caosdb_constants_COMPATIBLE_SERVER_VERSION_PRE_RELEASE() {
  return caosdb::COMPATIBLE_SERVER_VERSION_PRE_RELEASE;
}

int caosdb_status_code_OTHER_CLIENT_ERROR() { return caosdb::StatusCode::OTHER_CLIENT_ERROR; }

const char *caosdb_utility_get_env_var(const char *name, const char *fallback) {
  return caosdb::utility::get_env_var(name, fallback);
}

const char *caosdb_get_status_description(int code) {
  return caosdb::get_status_description(code).c_str();
}

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_create_pem_file_certificate_provider(
                    caosdb_connection_certificate_provider *out, const char *path),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    out->wrapped_certificate_provider =
                      new caosdb::configuration::PemFileCertificateProvider(std::string(path));
                    out->_deletable = true;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_delete_certificate_provider(
                    caosdb_connection_certificate_provider *provider),
                  {
                    if (provider->_deletable) {
                      delete static_cast<caosdb::configuration::CertificateProvider *>(
                        provider->wrapped_certificate_provider);
                    }
                    provider->_deletable = false;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_authentication_create_plain_password_authenticator(
                    caosdb_authentication_authenticator *out, const char *username,
                    const char *password),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    out->wrapped_authenticator =
                      new caosdb::authentication::PlainPasswordAuthenticator(std::string(username),
                                                                             std::string(password));
                    out->_deletable = true;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_authentication_delete_authenticator(
                    caosdb_authentication_authenticator *authenticator),
                  {
                    if (authenticator->_deletable) {
                      delete static_cast<caosdb::authentication::Authenticator *>(
                        authenticator->wrapped_authenticator);
                    }
                    authenticator->_deletable = false;
                    return 0;
                  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_connection_create_tls_connection_configuration(
    caosdb_connection_connection_configuration *out, const char *host, const int port,
    caosdb_authentication_authenticator *authenticator,
    caosdb_connection_certificate_provider *provider),
  {
    if (out->_deletable) {
      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
    }
    auto host_str = std::string(host);
    if (authenticator != nullptr && provider != nullptr) {
      auto wrapped_provider = static_cast<caosdb::configuration::CertificateProvider *>(
        provider->wrapped_certificate_provider);
      auto wrapped_authenticator =
        static_cast<caosdb::authentication::Authenticator *>(authenticator->wrapped_authenticator);
      out->wrapped_connection_configuration = new caosdb::configuration::TlsConnectionConfiguration(
        host_str, port, *wrapped_provider, *wrapped_authenticator);
    } else if (authenticator != nullptr) {
      auto wrapped_authenticator =
        static_cast<caosdb::authentication::Authenticator *>(authenticator->wrapped_authenticator);
      out->wrapped_connection_configuration = new caosdb::configuration::TlsConnectionConfiguration(
        host_str, port, *wrapped_authenticator);
    } else if (provider != nullptr) {
      auto wrapped_provider = static_cast<caosdb::configuration::CertificateProvider *>(
        provider->wrapped_certificate_provider);
      out->wrapped_connection_configuration =
        new caosdb::configuration::TlsConnectionConfiguration(host_str, port, *wrapped_provider);
    } else {
      out->wrapped_connection_configuration =
        new caosdb::configuration::TlsConnectionConfiguration(host_str, port);
    }
    out->_deletable = true;
    return 0;
  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_create_insecure_connection_configuration(
                    caosdb_connection_connection_configuration *out, const char *host,
                    const int port),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    out->wrapped_connection_configuration =
                      new caosdb::configuration::InsecureConnectionConfiguration(host, port);
                    out->_deletable = true;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_delete_connection_configuration(
                    caosdb_connection_connection_configuration *configuration),
                  {
                    if (configuration->_deletable) {
                      delete static_cast<caosdb::configuration::ConnectionConfiguration *>(
                        configuration->wrapped_connection_configuration);
                    }
                    configuration->_deletable = false;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_create_connection(
                    caosdb_connection_connection *out,
                    const caosdb_connection_connection_configuration *configuration),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    caosdb::configuration::ConnectionConfiguration *config =
                      static_cast<caosdb::configuration::ConnectionConfiguration *>(
                        configuration->wrapped_connection_configuration);
                    out->wrapped_connection = new caosdb::connection::Connection(*config);
                    out->_deletable = true;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_delete_connection(caosdb_connection_connection *connection),
                  {
                    if (connection->_deletable) {
                      delete static_cast<caosdb::connection::Connection *>(
                        connection->wrapped_connection);
                    }
                    connection->_deletable = false;
                    return 0;
                  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_connection_get_version_info(caosdb_info_version_info *out,
                                         const caosdb_connection_connection *connection),
  {
    auto *wrapped_connection =
      static_cast<caosdb::connection::Connection *>(connection->wrapped_connection);

    auto status = wrapped_connection->RetrieveVersionInfoNoExceptions();
    if (status.IsError()) {
      return status.GetCode();
    }
    auto version_info = wrapped_connection->GetVersionInfo();

    out->major = (int)version_info->GetMajor();
    out->minor = (int)version_info->GetMinor();
    out->patch = (int)version_info->GetPatch();

    // copy pre_release, needs local variable because out->pre_release is const
    auto *pre_release = (char *)malloc(sizeof(char) * (version_info->GetPreRelease().length() + 1));
    strcpy(pre_release, version_info->GetPreRelease().c_str());
    out->pre_release = pre_release;

    // copy build, needs local variable because out->build is const
    auto *build = (char *)malloc(sizeof(char) * (version_info->GetBuild().length() + 1));
    strcpy(build, version_info->GetBuild().c_str());
    out->build = build;

    return 0;
  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_connection_manager_get_default_connection(
                    caosdb_connection_connection *out),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    out->wrapped_connection =
                      caosdb::connection::ConnectionManager::GetDefaultConnection().get();
                    out->_deletable = false;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_connection_manager_get_connection(
                    caosdb_connection_connection *out, const char *name),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    out->wrapped_connection =
                      caosdb::connection::ConnectionManager::GetConnection(std::string(name)).get();
                    // managed by the connection manager now, so not
                    // to be deleted manually
                    out->_deletable = false;
                    return 0;
                  })

/****************************************************************************
 * ENTITY STUFF AND TRANSACTIONS
 ****************************************************************************/
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_connection_connection_create_transaction(
                    caosdb_connection_connection *connection, caosdb_transaction_transaction *out),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    auto *wrapped_connection =
                      static_cast<caosdb::connection::Connection *>(connection->wrapped_connection);
                    out->wrapped_transaction = wrapped_connection->CreateTransaction().release();
                    out->_deletable = true;
                    return 0;
                  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_transaction_delete_transaction(caosdb_transaction_transaction *transaction), {
    if (transaction->_deletable) {
      delete static_cast<caosdb::transaction::Transaction *>(transaction->wrapped_transaction);
    }
    return 0;
  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_transaction_retrieve_by_id(
                    caosdb_transaction_transaction *transaction, const char *id),
                  {
                    auto *wrapped_transaction = static_cast<caosdb::transaction::Transaction *>(
                      transaction->wrapped_transaction);
                    return wrapped_transaction->RetrieveById(std::string(id));
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_transaction_retrieve_and_download_file_by_id(
                    caosdb_transaction_transaction *transaction, const char *id, const char *path),
                  {
                    auto *wrapped_transaction = static_cast<caosdb::transaction::Transaction *>(
                      transaction->wrapped_transaction);
                    return wrapped_transaction->RetrieveAndDownloadFileById(std::string(id),
                                                                            std::string(path));
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_transaction_retrieve_by_ids(
                    caosdb_transaction_transaction *transaction, const char *ids[], int length),
                  {
                    auto *wrapped_transaction = static_cast<caosdb::transaction::Transaction *>(
                      transaction->wrapped_transaction);
                    return wrapped_transaction->RetrieveById(ids, ids + length);
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_transaction_query(
                    caosdb_transaction_transaction *transaction, const char *query),
                  {
                    auto *wrapped_transaction = static_cast<caosdb::transaction::Transaction *>(
                      transaction->wrapped_transaction);
                    return wrapped_transaction->Query(std::string(query));
                  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_transaction_transaction_insert_entity(caosdb_transaction_transaction *transaction,
                                                   caosdb_entity_entity *entity),
  {
    auto *wrapped_transaction =
      static_cast<caosdb::transaction::Transaction *>(transaction->wrapped_transaction);
    auto *wrapped_entity = static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);

    return wrapped_transaction->InsertEntity(wrapped_entity);
  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_transaction_transaction_update_entity(caosdb_transaction_transaction *transaction,
                                                   caosdb_entity_entity *entity),
  {
    auto *wrapped_transaction =
      static_cast<caosdb::transaction::Transaction *>(transaction->wrapped_transaction);
    auto *wrapped_entity = static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);

    return wrapped_transaction->UpdateEntity(wrapped_entity);
  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_transaction_delete_by_id(
                    caosdb_transaction_transaction *transaction, const char *id),
                  {
                    auto *wrapped_transaction = static_cast<caosdb::transaction::Transaction *>(
                      transaction->wrapped_transaction);

                    return wrapped_transaction->DeleteById(std::string(id));
                  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_transaction_transaction_execute(caosdb_transaction_transaction *transaction), {
    auto *wrapped_transaction =
      static_cast<caosdb::transaction::Transaction *>(transaction->wrapped_transaction);
    wrapped_transaction->ExecuteAsynchronously();
    auto status = wrapped_transaction->WaitForIt();
    return status.GetCode();
  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_transaction_transaction_get_result_set(caosdb_transaction_transaction *transaction,
                                                    caosdb_transaction_result_set *out),
  {
    if (out->_deletable) {
      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
    }
    auto *wrapped_transaction =
      static_cast<caosdb::transaction::Transaction *>(transaction->wrapped_transaction);
    out->wrapped_result_set = (void *)(&(wrapped_transaction->GetResultSet()));
    out->_deletable = false;
    return 0;
  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_transaction_get_count_result(
                    caosdb_transaction_transaction *transaction, long *out),
                  {
                    auto *wrapped_transaction = static_cast<caosdb::transaction::Transaction *>(
                      transaction->wrapped_transaction);
                    long cr(wrapped_transaction->GetCountResult());
                    *out = cr;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_result_set_at(caosdb_transaction_result_set *result_set,
                                                       caosdb_entity_entity *entity, int index),
                  {
                    auto *wrapped_result_set = static_cast<caosdb::transaction::MultiResultSet *>(
                      result_set->wrapped_result_set);
                    entity->wrapped_entity = wrapped_result_set->mutable_at(index);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_transaction_result_set_size(caosdb_transaction_result_set *result_set,
                                                         int *out),
                  {
                    auto *wrapped_result_set = static_cast<caosdb::transaction::MultiResultSet *>(
                      result_set->wrapped_result_set);
                    int size(wrapped_result_set->size());
                    *out = size;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR, int caosdb_entity_create_entity(caosdb_entity_entity *out), {
  if (out->_deletable) {
    return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
  }
  out->wrapped_entity = new caosdb::entity::Entity();
  out->_deletable = true;
  return 0;
})

ERROR_RETURN_CODE(GENERIC_ERROR, int caosdb_entity_delete_entity(caosdb_entity_entity *out), {
  if (out->_deletable) {
    delete static_cast<caosdb::entity::Entity *>(out->wrapped_entity);
  }
  out->_deletable = false;
  return 0;
})

ERROR_RETURN_CODE(GENERIC_ERROR, int caosdb_entity_create_property(caosdb_entity_property *out), {
  if (out->_deletable) {
    return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
  }
  out->wrapped_property = new caosdb::entity::Property();
  out->_deletable = true;
  return 0;
})

ERROR_RETURN_CODE(GENERIC_ERROR, int caosdb_entity_delete_property(caosdb_entity_property *out), {
  if (out->_deletable) {
    delete static_cast<caosdb::entity::Property *>(out->wrapped_property);
  }
  out->_deletable = false;
  return 0;
})

ERROR_RETURN_CODE(GENERIC_ERROR, int caosdb_entity_create_parent(caosdb_entity_parent *out), {
  if (out->_deletable) {
    return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
  }
  out->wrapped_parent = new caosdb::entity::Parent();
  out->_deletable = true;
  return 0;
})

ERROR_RETURN_CODE(GENERIC_ERROR, int caosdb_entity_delete_parent(caosdb_entity_parent *out), {
  if (out->_deletable) {
    delete static_cast<caosdb::entity::Parent *>(out->wrapped_parent);
  }
  out->_deletable = false;
  return 0;
})

CAOSDB_ENTITY_GET(id, GetId())
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_role(caosdb_entity_entity *entity, char **out), {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    std::string role_str = ENUM_NAME_FROM_VALUE(wrapped_entity->GetRole(), Role);
                    auto *tmp = (char *)malloc(sizeof(char) * role_str.length() + 1);
                    strcpy(tmp, role_str.c_str());
                    delete[] * out;
                    *out = tmp;
                    return 0;
                  })
CAOSDB_ENTITY_GET(name, GetName())
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_local_path(caosdb_entity_entity *entity, char **out),
                  {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    auto path = wrapped_entity->GetLocalPath().string();
                    auto *tmp = (char *)(malloc(sizeof(char) * path.length() + 1));
                    strcpy(tmp, path.c_str());
                    delete[] * out;
                    *out = tmp;
                    return 0;
                  })
// CAOSDB_ENTITY_GET(file_path, GetFilePath()) TODO(henrik)
CAOSDB_ENTITY_GET(description, GetDescription())
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_datatype(caosdb_entity_entity *entity, char **name,
                                                        bool *is_ref, bool *is_list),
                  {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    const auto &datatype = wrapped_entity->GetDataType();
                    *is_list = datatype.IsList();
                    std::string datatype_name;
                    if (*is_list) {
                      const auto &list_datatype = datatype.AsList();
                      *is_ref = list_datatype.IsListOfReference();
                      if (*is_ref) {
                        datatype_name = list_datatype.GetReferenceDataType().GetName();
                      } else {
                        datatype_name =
                          ENUM_NAME_FROM_VALUE(list_datatype.GetAtomicDataType(), AtomicDataType);
                      }
                    } else {
                      *is_ref = datatype.IsReference();
                      if (*is_ref) {
                        datatype_name = datatype.AsReference().GetName();
                      } else {
                        datatype_name = ENUM_NAME_FROM_VALUE(datatype.AsAtomic(), AtomicDataType);
                      }
                    }
                    char *tmp = (char *)malloc(sizeof(char) * datatype_name.length() + 1);
                    strcpy(tmp, datatype_name.c_str());
                    delete[] * name;
                    *name = tmp;
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_int_value(caosdb_entity_entity *entity, long *out), {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    *out = wrapped_entity->GetValue().AsInteger();
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_double_value(caosdb_entity_entity *entity,
                                                            double *out),
                  {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    *out = wrapped_entity->GetValue().AsDouble();
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_boolean_value(caosdb_entity_entity *entity,
                                                             bool *out),
                  {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    *out = wrapped_entity->GetValue().AsBool();
                    return 0;
                  })
CAOSDB_ENTITY_GET(string_value, GetValue().AsString())

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_int_list_value_at(caosdb_entity_entity *entity,
                                                                 long *out, const int index),
                  {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    auto value_list = wrapped_entity->GetValue().AsList();
                    *out = value_list[index].AsInteger();
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_double_list_value_at(caosdb_entity_entity *entity,
                                                                    double *out, const int index),
                  {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    auto value_list = wrapped_entity->GetValue().AsList();
                    *out = value_list[index].AsDouble();
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_boolean_value_list_value_at(
                    caosdb_entity_entity *entity, bool *out, const int index),
                  {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    auto value_list = wrapped_entity->GetValue().AsList();
                    *out = value_list[index].AsBool();
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_string_list_value_at(caosdb_entity_entity *entity,
                                                                    char **out, const int index),
                  {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    auto value_list = wrapped_entity->GetValue().AsList();
                    char *tmp =
                      (char *)malloc(sizeof(char) * value_list[index].AsString().length() + 1);
                    strcpy(tmp, value_list[index].AsString().c_str());
                    delete[] * out;
                    *out = tmp;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_value_list_length(caosdb_entity_entity *entity,
                                                                 int *out),
                  {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    if (wrapped_entity->GetValue().IsList()) {
                      *out = wrapped_entity->GetValue().AsList().size();
                    } else {
                      *out = 0;
                    }
                    return 0;
                  })

CAOSDB_ENTITY_GET(unit, GetUnit())
CAOSDB_ENTITY_GET(version_id, GetVersionId())

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_errors_size(caosdb_entity_entity *entity, int *out),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    *out = wrapped_entity->GetErrors().size();
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_error(caosdb_entity_entity *entity,
                                                     caosdb_entity_message *out, int index),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    out->wrapped_message = wrapped_entity->GetErrors().mutable_at(index);
                    out->_deletable = false;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_warnings_size(caosdb_entity_entity *entity,
                                                             int *out),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    *out = wrapped_entity->GetWarnings().size();
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_warning(caosdb_entity_entity *entity,
                                                       caosdb_entity_message *out, int index),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    out->wrapped_message = wrapped_entity->GetWarnings().mutable_at(index);
                    out->_deletable = false;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_infos_size(caosdb_entity_entity *entity, int *out), {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    *out = wrapped_entity->GetInfos().size();
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_info(caosdb_entity_entity *entity,
                                                    caosdb_entity_message *out, int index),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    out->wrapped_message = wrapped_entity->GetInfos().mutable_at(index);
                    out->_deletable = false;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_properties_size(caosdb_entity_entity *entity,
                                                               int *out),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    *out = wrapped_entity->GetProperties().size();
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_property(caosdb_entity_entity *entity,
                                                        caosdb_entity_property *out, int index),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    out->wrapped_property = wrapped_entity->GetProperties().mutable_at(index);
                    out->_deletable = false;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_parents_size(caosdb_entity_entity *entity, int *out),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    *out = wrapped_entity->GetParents().size();
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_get_parent(caosdb_entity_entity *entity,
                                                      caosdb_entity_parent *out, int index),
                  {
                    if (out->_deletable) {
                      return caosdb::StatusCode::EXTERN_C_ASSIGNMENT_ERROR;
                    }
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    out->wrapped_parent = wrapped_entity->GetParents().mutable_at(index);
                    out->_deletable = false;
                    return 0;
                  })

CAOSDB_PARENT_GET(id, GetId())
CAOSDB_PARENT_GET(name, GetName())
CAOSDB_PARENT_GET(description, GetDescription())

CAOSDB_PROPERTY_GET(id, GetId())
CAOSDB_PROPERTY_GET(name, GetName())
CAOSDB_PROPERTY_GET(description, GetDescription())

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_get_importance(caosdb_entity_property *property,
                                                            char **out),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    std::string importance_str =
                      ENUM_NAME_FROM_VALUE(wrapped_property->GetImportance(), Importance);
                    char *tmp = (char *)malloc(sizeof(char) * importance_str.length() + 1);
                    strcpy(tmp, importance_str.c_str());
                    delete[] * out;
                    *out = tmp;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_get_datatype(caosdb_entity_property *property,
                                                          char **name, bool *is_ref, bool *is_list),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    const auto &datatype = wrapped_property->GetDataType();
                    *is_list = datatype.IsList();
                    std::string datatype_name;
                    if (*is_list) {
                      const auto &list_datatype = datatype.AsList();
                      *is_ref = list_datatype.IsListOfReference();
                      if (*is_ref) {
                        datatype_name = list_datatype.GetReferenceDataType().GetName();
                      } else {
                        datatype_name =
                          ENUM_NAME_FROM_VALUE(list_datatype.GetAtomicDataType(), AtomicDataType);
                      }
                    } else {
                      *is_ref = datatype.IsReference();
                      if (*is_ref) {
                        datatype_name = datatype.AsReference().GetName();
                      } else {
                        datatype_name = ENUM_NAME_FROM_VALUE(datatype.AsAtomic(), AtomicDataType);
                      }
                    }
                    char *tmp = (char *)malloc(sizeof(char) * datatype_name.length() + 1);
                    strcpy(tmp, datatype_name.c_str());
                    delete[] * name;
                    *name = tmp;
                    return 0;
                  })

CAOSDB_PROPERTY_GET(unit, GetUnit())

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_get_int_value(caosdb_entity_property *property,
                                                           long *out),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    *out = wrapped_property->GetValue().AsInteger();
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_get_double_value(caosdb_entity_property *property,
                                                              double *out),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    *out = wrapped_property->GetValue().AsDouble();
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_get_boolean_value(caosdb_entity_property *property,
                                                               bool *out),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    *out = wrapped_property->GetValue().AsBool();
                    return 0;
                  })
CAOSDB_PROPERTY_GET(string_value, GetValue().AsString())

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_get_int_list_value_at(caosdb_entity_property *property,
                                                                   long *out, const int index),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    auto value_list = wrapped_property->GetValue().AsList();
                    *out = value_list[index].AsInteger();
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_get_double_list_value_at(
                    caosdb_entity_property *property, double *out, const int index),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    auto value_list = wrapped_property->GetValue().AsList();
                    *out = value_list[index].AsDouble();
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_get_boolean_value_list_value_at(
                    caosdb_entity_property *property, bool *out, const int index),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    auto value_list = wrapped_property->GetValue().AsList();
                    *out = value_list[index].AsBool();
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_get_string_list_value_at(
                    caosdb_entity_property *property, char **out, const int index),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    auto value_list = wrapped_property->GetValue().AsList();
                    auto *tmp =
                      (char *)malloc(sizeof(char) * value_list[index].AsString().length() + 1);
                    strcpy(tmp, value_list[index].AsString().c_str());
                    delete[] * out;
                    *out = tmp;
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_get_value_list_length(caosdb_entity_property *property,
                                                                   int *out),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    if (wrapped_property->GetValue().IsList()) {
                      *out = wrapped_property->GetValue().AsList().size();
                    } else {
                      *out = 0;
                    }
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_message_get_code(caosdb_entity_message *message, int *out), {
                    auto *wrapped_message =
                      static_cast<caosdb::entity::Message *>(message->wrapped_message);
                    *out = wrapped_message->GetCode();
                    return 0;
                  })

ERROR_RETURN_CODE(
  GENERIC_ERROR,
  int caosdb_entity_message_get_description(caosdb_entity_message *message, char **out), {
    auto *wrapped_message = static_cast<caosdb::entity::Message *>(message->wrapped_message);
    auto *tmp = (char *)malloc(sizeof(char) * wrapped_message->GetDescription().length() + 1);
    strcpy(tmp, wrapped_message->GetDescription().c_str());
    delete[] * out;
    *out = tmp;
    return 0;
  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_set_role(caosdb_entity_entity *entity, const char *role),
                  {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    try {
                      auto enum_value = ENUM_VALUE_FROM_NAME(std::string(role), Role);
                      wrapped_entity->SetRole(enum_value);
                      return 0;
                    } catch (const std::out_of_range &exc) {
                      caosdb::logging::caosdb_log_fatal(CCAOSDB_LOGGER_NAME, exc.what());
                      return caosdb::StatusCode::ENUM_MAPPING_ERROR;
                    }
                  })
CAOSDB_ENTITY_SET(name, name, wrapped_entity->SetName(std::string(name));)
CAOSDB_ENTITY_SET(local_path, local_path,
                  return wrapped_entity->SetLocalPath(boost::filesystem::path(local_path));)
CAOSDB_ENTITY_SET(file_path, file_path, wrapped_entity->SetFilePath(std::string(file_path));)
CAOSDB_ENTITY_SET(description, description,
                  wrapped_entity->SetDescription(std::string(description));)
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_set_datatype(caosdb_entity_entity *entity,
                                                        const char *datatype, const bool is_ref,
                                                        const bool is_list),
                  {
                    auto *wrapped_entity = WRAPPED_ENTITY_CAST(entity);
                    if (is_ref) {
                      // Refernce datatype with name of reference
                      wrapped_entity->SetDataType(std::string(datatype), is_list);
                      return 0;
                    } else {
                      // Atomic datatype so get from enum
                      try {
                        auto enum_value =
                          ENUM_VALUE_FROM_NAME(std::string(datatype), AtomicDataType);
                        wrapped_entity->SetDataType(enum_value, is_list);
                        return 0;
                      } catch (const std::out_of_range &exc) {
                        caosdb::logging::caosdb_log_fatal(CCAOSDB_LOGGER_NAME, exc.what());
                        return caosdb::StatusCode::ENUM_MAPPING_ERROR;
                      }
                    }
                  })
CAOSDB_ENTITY_SET(unit, unit, wrapped_entity->SetUnit(std::string(unit));)

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_set_int_value(caosdb_entity_entity *entity,
                                                         const long value),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    wrapped_entity->SetValue(value);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_set_double_value(caosdb_entity_entity *entity,
                                                            const double value),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    wrapped_entity->SetValue(value);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_set_boolean_value(caosdb_entity_entity *entity,
                                                             const bool value),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    wrapped_entity->SetValue(value);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_set_string_value(caosdb_entity_entity *entity,
                                                            const char *value),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    wrapped_entity->SetValue(std::string(value));
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_set_int_list_value(caosdb_entity_entity *entity,
                                                              const long *value, const int length),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    std::vector<long> value_list;
                    for (int i = 0; i < length; i++) {
                      value_list.push_back(value[i]);
                    }
                    wrapped_entity->SetValue(value_list);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_set_double_list_value(caosdb_entity_entity *entity,
                                                                 const double *value,
                                                                 const int length),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    std::vector<double> value_list;
                    for (int i = 0; i < length; i++) {
                      value_list.push_back(value[i]);
                    }
                    wrapped_entity->SetValue(value_list);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_set_boolean_list_value(caosdb_entity_entity *entity,
                                                                  const bool *value,
                                                                  const int length),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    std::vector<bool> value_list;
                    for (int i = 0; i < length; i++) {
                      value_list.push_back(value[i]);
                    }
                    wrapped_entity->SetValue(value_list);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_set_string_list_value(caosdb_entity_entity *entity,
                                                                 const char **value,
                                                                 const int length),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    std::vector<std::string> value_list;
                    for (int i = 0; i < length; i++) {
                      value_list.push_back(std::string(value[i]));
                    }
                    wrapped_entity->SetValue(value_list);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_append_parent(caosdb_entity_entity *entity,
                                                         caosdb_entity_parent *parent),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    auto *wrapped_parent =
                      static_cast<caosdb::entity::Parent *>(parent->wrapped_parent);
                    wrapped_entity->AppendParent(*wrapped_parent);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_remove_parent(caosdb_entity_entity *entity, int index), {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    wrapped_entity->RemoveParent(index);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_append_property(caosdb_entity_entity *entity,
                                                           caosdb_entity_property *property),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    auto *wrapped_property =
                      static_cast<caosdb::entity::Property *>(property->wrapped_property);
                    wrapped_entity->AppendProperty(*wrapped_property);
                    return 0;
                  })
ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_entity_remove_property(caosdb_entity_entity *entity, int index),
                  {
                    auto *wrapped_entity =
                      static_cast<caosdb::entity::Entity *>(entity->wrapped_entity);
                    wrapped_entity->RemoveProperty(index);
                    return 0;
                  })

CAOSDB_PARENT_SET(id, id, wrapped_parent->SetId(std::string(id));)
CAOSDB_PARENT_SET(name, name, wrapped_parent->SetName(std::string(name));)

CAOSDB_PROPERTY_SET(name, name, wrapped_property->SetName(std::string(name));)
CAOSDB_PROPERTY_SET(id, id, wrapped_property->SetId(std::string(id));)

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_set_datatype(caosdb_entity_property *property,
                                                          const char *datatype, const bool is_ref,
                                                          const bool is_list),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    if (is_ref) {
                      // Refernce datatype with name of reference
                      wrapped_property->SetDataType(std::string(datatype), is_list);
                      return 0;
                    } else {
                      // Atomic datatype so get from enum
                      try {
                        auto enum_value =
                          ENUM_VALUE_FROM_NAME(std::string(datatype), AtomicDataType);
                        wrapped_property->SetDataType(enum_value, is_list);
                        return 0;
                      } catch (const std::out_of_range &exc) {
                        caosdb::logging::caosdb_log_fatal(CCAOSDB_LOGGER_NAME, exc.what());
                        return caosdb::StatusCode::ENUM_MAPPING_ERROR;
                      }
                    }
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_set_importance(caosdb_entity_property *property,
                                                            const char *importance),
                  {
                    auto *wrapped_property = WRAPPED_PROPERTY_CAST(property);
                    try {
                      auto enum_value = ENUM_VALUE_FROM_NAME(std::string(importance), Importance);
                      wrapped_property->SetImportance(enum_value);
                      return 0;
                    } catch (const std::out_of_range &exc) {
                      caosdb::logging::caosdb_log_fatal(CCAOSDB_LOGGER_NAME, exc.what());
                      return caosdb::StatusCode::ENUM_MAPPING_ERROR;
                    }
                  })

CAOSDB_PROPERTY_SET(unit, unit, wrapped_property->SetUnit(std::string(unit));)

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_set_int_value(caosdb_entity_property *property,
                                                           const long value),
                  {
                    auto *wrapped_property =
                      static_cast<caosdb::entity::Property *>(property->wrapped_property);
                    wrapped_property->SetValue(value);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_set_double_value(caosdb_entity_property *property,
                                                              const double value),
                  {
                    auto *wrapped_property =
                      static_cast<caosdb::entity::Property *>(property->wrapped_property);
                    wrapped_property->SetValue(value);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_set_boolean_value(caosdb_entity_property *property,
                                                               const bool value),
                  {
                    auto *wrapped_property =
                      static_cast<caosdb::entity::Property *>(property->wrapped_property);
                    wrapped_property->SetValue(value);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_set_string_value(caosdb_entity_property *property,
                                                              const char *value),
                  {
                    auto *wrapped_property =
                      static_cast<caosdb::entity::Property *>(property->wrapped_property);
                    wrapped_property->SetValue(std::string(value));
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_set_int_list_value(caosdb_entity_property *property,
                                                                const long *value,
                                                                const int length),
                  {
                    auto *wrapped_property =
                      static_cast<caosdb::entity::Property *>(property->wrapped_property);
                    std::vector<long> value_list;
                    for (int i = 0; i < length; i++) {
                      value_list.push_back(value[i]);
                    }
                    wrapped_property->SetValue(value_list);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_set_double_list_value(caosdb_entity_property *property,
                                                                   const double *value,
                                                                   const int length),
                  {
                    auto *wrapped_property =
                      static_cast<caosdb::entity::Property *>(property->wrapped_property);
                    std::vector<double> value_list;
                    for (int i = 0; i < length; i++) {
                      value_list.push_back(value[i]);
                    }
                    wrapped_property->SetValue(value_list);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_set_boolean_list_value(
                    caosdb_entity_property *property, const bool *value, const int length),
                  {
                    auto *wrapped_property =
                      static_cast<caosdb::entity::Property *>(property->wrapped_property);
                    std::vector<bool> value_list;
                    for (int i = 0; i < length; i++) {
                      value_list.push_back(value[i]);
                    }
                    wrapped_property->SetValue(value_list);
                    return 0;
                  })

ERROR_RETURN_CODE(GENERIC_ERROR,
                  int caosdb_entity_property_set_string_list_value(caosdb_entity_property *property,
                                                                   const char **value,
                                                                   const int length),
                  {
                    auto *wrapped_property =
                      static_cast<caosdb::entity::Property *>(property->wrapped_property);
                    std::vector<std::string> value_list;
                    for (int i = 0; i < length; i++) {
                      value_list.push_back(std::string(value[i]));
                    }
                    wrapped_property->SetValue(value_list);
                    return 0;
                  })
}
